@@ -8,6 +8,7 @@ from __future__ import unicode_literals
 import ast
 import codecs
 import sys
+from io import open
 
 import gdb
 
@@ -194,7 +195,7 @@ theme.Parameter('code-prefix', '►', "prefix marker for 'context code' command"
 @pwndbg.memoize.reset_on_start
 def get_highlight_source(filename):
     # Notice that the code is cached
-    with open(filename) as f:
+    with open(filename, encoding='utf-8') as f:
         source = f.read()
 
     if pwndbg.config.syntax_highlight:
@@ -260,7 +261,7 @@ def context_code():
             )
             formatted_source.append(line)
 
-        banner = [pwndbg.ui.banner("Source (code)")]
+        banner = [pwndbg.ui.banner("Source (code)"), 'In file: %s' % filename]
         banner.extend(formatted_source)
         return banner
     except:
